/*!
 * Copyright (C) 2005 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://jtagomat.sourceforge.net/
 */

/*!
 * \file comm.c
 * \brief Turtelizer protocol implementation.
 *
 * \verbatim
 *
 * $Log: comm.c,v $
 * Revision 1.1  2006/01/09 09:17:40  haraldkipp
 * First check-in.
 *
 *
 * \endverbatim
 */

#include <avr/io.h>
#include <avr/boot.h>
#include <avr/pgmspace.h>

#include "led.h"
#include "comm.h"

#define STX     0x02
#define ETX     0x03
#define ACK     0x06
#define DLE     0x10
#define NAK     0x15

unsigned char tbuf[MAX_TELESIZE];
unsigned char tlen;

/*!
 * \brief Enable UART communication.
 */
void InitUart(void)
{
    unsigned short cnt;

    UBRRL = 3;
    UCSRB = (1 << RXEN) | (1 << TXEN);

    for (cnt = 0; cnt; cnt++) {
        asm volatile("nop");
    }
}

static void SendChar(unsigned char ch)
{
    while((UCSRA & _BV(UDRE)) == 0);
    UDR = ch;
}

/*!
 * \brief Receive an STX/ETX/CKS telegram.
 *
 * \param buf Pointer to the buffer that receives the pure data.
 * \param siz 
 */
void RecvTelegram(void)
{
    unsigned short blink = 0;
    unsigned char ch;
    unsigned char lrc = 0;
    unsigned char st = 0;

    for (;;) {
        /*
         * Wait until character available and keep the
         * LED blinking.
         */
        while((UCSRA & _BV(RXC)) == 0) {
            if (blink++ == 0) {
                Led(1);
            }
            else if (blink == 0x4000) {
                Led(0);
            }
        }

        /* Get the character and update the checksum. */
        ch = UDR;
        lrc ^= ch;

        /*
         * In state 1 we are receiving telegram data.
         */
        if (st == 1) {
            /* An un-escaped STX, fall back to state 0. */
            if (ch == STX) {
                st = 0;
            }
            /* An escape, switch to state 2. */
            else if (ch == DLE) {
                st = 2;
                lrc ^= ch;
                continue;
            }
            /* ETX marks end of data, switch to state 3. */
            else if (ch == ETX) {
                st = 3;
                continue;
            }
            /* At this point we are either in state 0 or 1. */
        }

        /*
         * In state 0 we are waiting for the telegram start.
         */
        if (st == 0) {
            /* Got it! Clear the checksum and the receive counter. */
            if (ch == STX) {
                lrc = 0;
                tlen = 0;
                st++;
            }
            /* The user may press Ctrl-V to query our version code. */
            else if (ch == 22) {
                char *vm = "TurtleBoot " VERSION_STRING "\r\n";
                while (*vm) {
                    SendChar(*vm++);
                }
            }
            /* Anything else is ignored in state 0. */
            continue;
        }

        /* In state 3 we expect the longitudinal checksum. */
        if (st == 3) {
            /* 
             * Ex-oring all characters must clear all bits. Otherwise 
             * the checksum was bad and we send a NAK response. A NAK
             * is also send in case of a receive buffer overflow.
             */
            if (lrc || tlen > MAX_TELESIZE) {
                st = 0;
                SendChar(NAK);
                continue;
            }
            SendChar(ACK);
            break;
        }

        /* At this point we got a data byte, either escaped (state 2)
           or un-escaped (state 1). Check for buffer overflow. */
        if (tlen < MAX_TELESIZE) {
            tbuf[tlen] = ch;
        }
        tlen++;
        st = 1;
    }
}

/*!
 * \brief Send an STX/ETX/CKS telegram.
 */
void SendTelegram(void)
{
    unsigned char i;
    unsigned char ch;
    unsigned char lrc;
    unsigned char retry;

    for (retry = 0; retry < 3; retry++) {
        /*
         * Send STX DATA ETX LRC.
         */
        lrc = 0;
        SendChar(STX);
        for (i = 0; i < tlen; i++) {
            ch = tbuf[i];
            lrc ^= ch;
            if (ch == STX || ch == ETX || ch == DLE) {
                SendChar(DLE);
            }
            SendChar(ch);
        }
        SendChar(ETX);
        lrc ^= ETX;
        SendChar(lrc);

        /* Wait for ACK, maintain short timeout. */
        ch = NAK;
        for (i = 1; i; i++) {
            if(UCSRA & _BV(RXC)) {
                ch = UDR;
                break;
            }
        }

        /*
         * We do retries only on NAK or timeout, Either this
         * was an ACK or the partner tries to send a new telegram.
         */
        if (ch != NAK) {
            break;
        }
    }
}

